package org.unisb.dbs.util;

import java.util.*;
import java.io.Serializable;

public class SparseVector implements Serializable {
  int size;
  int[] keys;
  double[] values;
  
  public SparseVector() {
    this(1);
  }
  
  public SparseVector(int capacity) {
    keys = new int[capacity];
    values = new double[capacity];
    Arrays.fill(keys, 0, capacity, Integer.MAX_VALUE);
    size = 0;
  }
  
  /**
   * The number of used dimensions.
   * <P>
   * This is the total number of non-zero dimensions. It is not equal to the
   * number of the highest indexed dimension.
   *
   * @return the number of non-zero dimensions
   */
  public int size() {
    return size;
  }
  
  /**
   * Set the value at a particular dimension.
   * <P>
   * This silently handles the case when previously there was no value at dim.
   * It does not contract the sparse array if you set a value to zero.
   *
   * @param dim  the dimension to alter
   * @param value the new value
   */
  public void put(int dim, double value) {
    // find index of key nearest dim
    int indx = Arrays.binarySearch(keys, dim);
    
    if(indx >= 0) { // found entry for dim
      values[indx] = value;
    } else { // need to create entry for dim
      indx = -(indx + 1);
      
      if ((size + 1) >= keys.length) { // growing arrays
        int[] nKeys = new int[keys.length * 2];
        Arrays.fill(nKeys, size, nKeys.length, Integer.MAX_VALUE);
        System.arraycopy(keys, 0, nKeys, 0, indx);
        System.arraycopy(keys, indx, nKeys, indx+1, size-indx);
        keys = nKeys;
        
        double[] nValues = new double[values.length * 2];
        Arrays.fill(nValues, size, nValues.length, Double.NaN);
        System.arraycopy(values, 0, nValues, 0, indx);
        System.arraycopy(values, indx, nValues, indx+1, size-indx);
        values = nValues;
      } else {
        try {
          System.arraycopy(keys, indx, keys, indx+1, size-indx);
          System.arraycopy(values, indx, values, indx+1, size-indx);
        } catch (ArrayIndexOutOfBoundsException ae) {
          // paranoya check - they were out to get me once
          //                - they may try it again
          System.out.println("dim         = " + dim);
          System.out.println("value       = " + value);
          System.out.println("keys.length = " + keys.length);
          System.out.println("size        = " + size);
          System.out.println("indx        = " + indx);
          System.out.println("indx+1      = " + (indx+1));
          System.out.println("size-indx   = " + (size-indx));
          System.out.println("size+1      = " + (size+1));
          throw ae;
        }
      }
      
      keys[indx] = dim;
      values[indx] = value;
      ++size;
    }
  }
  
  /**
   * Retrieve the value at dimension dim.
   *
   * @param dim the dimension to retrieve a value for
   * @return the value at that dimension
   */
  public double get(int dim) {
    int pos = Arrays.binarySearch(keys, dim);
    if (pos >= 0) {
      return values[pos];
    }
    return 0.0;
  }
  
  /**
   * Retrieve the dimension at a specific index.
   * <P>
   * E.g., if the sparse vector had dimensions 5, 11, 12, 155 set to non-zero
   * values then <code>sv.getDimAtIndex(2)</code> would return 12.
   *
   * @param indx  the index
   * @return the dimension stoored at that index
   * @throws ArrayIndexOutOfBoundsException if index >= size.
   */
  public int getDimAtIndex(int indx) {
    if(indx >= size) {
      throw new ArrayIndexOutOfBoundsException(
      "Attempted to read item " + indx + " of an array with " + size + "elements"
      );
    }
    return keys[indx];
  }
  
  /**
   * Retrieve the value at a specific index.
   * <P>
   * E.g., if the sparse vector contained the data 5->0.1, 11->100, 12->8.5, 155->-10
   * then <code>sv.geValueAtIndex(2)</code> would return 8.5.
   *
   * @param indx  the index
   * @return the value stoored at that index
   * @throws ArrayIndexOutOfBoundsException if index >= size.
   */
  public double getValueAtIndex(int indx) {
    if(indx >= size) {
      throw new ArrayIndexOutOfBoundsException(
      "Attempted to read item " + indx + " of an array with " + size + "elements"
      );
    }
    return values[indx];
  }
  
  public int maxIndex() {
    return keys[size - 1];
  }
  
  public static SparseVector normalLengthVector(SparseVector v, double length) {
    SparseVector n = new SparseVector(v.size);
    
    double oldLength = 0;
    for (int i = 0; i < v.size; ++i) {
      oldLength += v.values[i] * v.values[i];
    }
    oldLength = Math.sqrt(oldLength);
    
    for (int i = 0; i < v.size; ++i) {
      n.put(v.keys[i], v.values[i] * length / oldLength);
    }
    
    return n;
  }
  
  public void normTF() {
    double norm = getValueSum();
    for ( int i=0; i < values.length; i++ ) {
      if ( values[i] > 0 ) {
        values[i] /= norm;
      }
    }
  }
  
  
  
  public void normL2() {
    double norm = getNormL2();
    for ( int i=0; i < values.length; i++ ) {
      if ( values[i] > 0 ) {
        //values[i] /= norm;
        values[i] = Math.pow( values[i], 2 ) / Math.pow( norm, 2 );
      }
    }
  }
  
  public double getNormL2() {
    int i;
    double sum = 0.0;
    
    for ( i=0; i < values.length; i++ ) {
      if ( values[i] > 0 ) {
        sum += ( values[i] * values[i] );
      }
    }
    
    return Math.sqrt( sum );
  }
  
  public double getValueSum() {
    int i;
    double sum = 0.0;
    
    for ( i=0; i < values.length; i++ ) {
      if ( values[i] > 0 ) {
        sum += values[i];
      }
    }
    
    return sum;
  }
  
  public double cosine( SparseVector document ) {
    double sim = 0.0;
    for ( int i = 0; i < document.size(); i++ ) {
      int dim = document.getDimAtIndex( i );
      double qi = document.get( dim );
      double di = this.get( dim );
      sim += qi * di;
    }
    return sim / ( document.getNormL2() * this.getNormL2() );
  }

  public String toString() {
    String s = new String();
    s += "[ ";
    for ( int i = 0; i < size(); i++ ) {
      s += String.valueOf( getDimAtIndex( i ) ) + ": " + String.valueOf( getValueAtIndex( i ) );
      if ( i < size()-1 ) s += ", ";
    }
    s += " ]";
    return s;
  }
}
